need("mbase")
need("mldf")
need("mgridio")

--- Lua Wrapper for gridgen module
--- 
--- \file wrap_gridgen.lua
--- \author J.Fuhrmann
---
---
--- The Lua language has more flexibility in setting up arrays.
--- This module organizes the translation of a Lua table containing
--- the necessary data into a \link GridGenIO GridGenIO \endlink structure.
--- The general flow of information is:
--- <pre>
--- Lua like description table (.geom file)
---         |
---         |      GridGenIO()
---         v
--- Input GridGenIO 
---         |
---         |      tetgen()/triangle()/gridgen1D() 
---         v
--- Output GridGenIO 
---         |
---         |      gCreateGG()
---         v
---       Grid
--- </pre>

------------------------------------------------
---
--- \brief Main GridGenIO constructor.
---
--- Construct GridGenIO from description table.
---
--+ GridGenIO *GridGenIO(table);


function ggNew(desc)
   -- Pass through

   if istype(desc,"GridGenIO") then return desc end
   
   if desc.class=="gdesc" then desc:prepare_gridgenio() end
   
   local flags="pqAaV"   
   if desc.flags then flags=desc.flags end

   if desc.file then
     -- TODO: read many formats
      local stub,ext=splitfilename(desc.file)
      if ext=="poly" or ext=="smesh" then
        -- TODO check dim
        return mgridio.ggLoadTetgen(fname)
      else
	print("file extension "..ext.."not supported")
      end	
   end
   
   local gg=mgridio.ggCreate()
   
   if bool(desc.points) then gg.points=DArray2(desc.points) end
   local dim=mbase.da2NCol(gg.points)
   
   if desc.debug and dim == 2 then 
      print("\nfacet,region, p1,p2,coord1,coord2:")
      for i=1,table.getn( desc.facets) do
	 v=desc.facets[i]
	 print(i,desc.facetregions[i],v[1],v[2], desc.points[v[1]][1],desc.points[v[1]][2],desc.points[v[2]][1],desc.points[v[2]][2])
      end
      print("\nregion, number, coord, vol:")
      for i=1, table.getn( desc.regionnumbers) do
	 print(i,desc.regionnumbers[i],desc.regionpoints[i][1],desc.regionpoints[i][2],desc.regionvolumes[i])
      end
   end
   if bool(desc.cells) then gg.cells=IArray2(desc.cells) end
   if bool(desc.cellregions) then gg.cellregions=IArray1(desc.cellregions) end
   if bool(desc.cellvolumes) then gg.cellvolumes=DArray1(desc.cellvolumes) end
   
   if bool(desc.facetpolygonlist) then
      if bool(desc.facetregions) then gg.facetregions= IArray1(desc.facetregions) end
      if bool(desc.facetpolygonlist) then gg.facetpolygonlist= IArray1(desc.facetpolygonlist) end
      if bool(desc.facetpolygonindexes) then gg.facetpolygonindexes= IArray1(desc.facetpolygonindexes) end
      if bool(desc.facetpolygons) then gg.facetpolygons= IArray1(desc.facetpolygons) end
      if bool(desc.facetholes) then gg.facetholes=IArray1(desc.facetholes) end
      if bool(desc.facetholepoints) then gg.facetholepoints= DArray2(desc.facetholepoints) end
   elseif bool(desc.facets) and type(desc.facets)=="table" and desc.facets[1].polygons then
      genfacets_f(gg,desc.facets)
   elseif bool(desc.facets) then 
      -- detect if we deal with a table of facets of different size
      if dim==3 and type(desc.facets)=="table" then
	 local homog=0 -- TODO: so reicht es nicht, da z.B. im Fall 'nur Rechtecke' nicht genfacets_s() gerufen wird.
	 local l=table.getn(desc.facets[1])
	 for i=2,table.getn(desc.facets) do
	    if table.getn(desc.facets[i])~=l then
	       homog=0
	       break
	    end
	 end
	 if homog==1 then
	    gg.facets=IArray2(desc.facets) 
	 else
	    genfacets_s(gg,desc.facets)
	 end
      else -- 2D  array on input
	 gg.facets=IArray2(desc.facets)
      end
   end

   if bool(desc.facetregions) then 
     gg.facetregions=IArray1(desc.facetregions)
  elseif bool(desc.facets) then
     gg.facetregions=IArray1{n=desc.facets.nrow,init=1}
  end
  --  print(desc.regionvolumes)
  -- print("wrap_ggen:",DArray1)
   if bool(desc.regionpoints) then gg.regionpoints=DArray2(desc.regionpoints) end
   if bool(desc.regionnumbers) then gg.regionnumbers=IArray1(desc.regionnumbers) end
   if bool(desc.regionvolumes) then gg.regionvolumes=Array(desc.regionvolumes) end
   if bool(desc.eps) then gg=mgridio.ggIdentifyPoints(gg,desc.eps) end
   local ext={  flags=desc.flags,
                nproc=desc.nproc,
                cache=desc.cache,
                maxlen=desc.maxlen,
                hmax=desc.hmax}
   return gg,ext
end


---
--- \brief Allow to call gridgen1d with Lua table arguments
---
---
--+ GridGenIO *gridgen1d(table,flags);

function gridgen1d(gg,r)
  local gio,ext = ggNew(gg)
  if ext and ext.maxlen then r=ext.maxlen end
  return mgridio.gridgen1d(gio,r)
end

---
--- \brief Allow to call triangle with Lua table arguments
---
---
--+ GridGenIO *triangle(table,flags);

function triangle(gg,flags,...)
   local gio=gg
   local r=flags
   local ext=arg[1]
   if not istype(desc,"GridGenIO") then
      gio,xext = ggNew(gg)
      if xext then ext=xext end
   end
   if ext and ext.hmax then 
      -- set global refinement function to be called if -u is given
      _triref=ext.hmax
   end
   if _triref then
      local triref=ggLuaEdgeRefCreate("_triref")
      gg_out= mgridio.ggTriangle(gio,r,triref)
   else
      gg_out= mgridio.ggTriangle(gio,r)
   end
   _triref=nil
   return gg_out
end




---
--- \brief Allow to call tetgen with Lua table arguments
---
---
--+ GridGenIO *tetgen(table,flags);
function tetgen(gg,r)
   local gio,ext = ggNew(gg)
   local gg_out
   if ext and ext.flags then 
      r=ext.flags 
      -- set global refinement function to be called if -u is given
      _tetref=ext.hmax
   end
   if _tetref then
      local tetref=ggLuaEdgeRefCreate("_tetref")
      gg_out= mgridio.ggTetgen(gio,r,tetref)
   else
      gg_out= mgridio.ggTetgen(gio,r)
   end
  _tetref=nil
  return gg_out
end


------------------------------------------------
--- \luaonly GridGenIO class extension
---
--+  class  GridGenIO: public  GridGenIO{
--+    public:
--+ 
---  View grid. 
---  \see ggView()
--+ 
--+   view(glViewer *viewer); 
--+
---   Constructor. 
---
---   Instead of GridGenIO.new(desc), you can call GridGenIO(desc).
--+
--+   new(table desc); 
--+ }

GridGenIO=ggNew

--[[
mgridio.GridGenIO.view=ggView
mgridio.GridGenIO.new=ggNew

GridGenIO.__index_old=GridGenIO.__index;
GridGenIO.__newindex_old=GridGenIO.__newindex;

function GridGenIO.__newindex(v,idx,val)
   if type(val)=="userdata" then val.refcount=val.refcount+1 end
   GridGenIO.__newindex_old(v,idx,val)
end

function GridGenIO.__index(v,idx) 
      return luaref(GridGenIO.__index_old(v,idx))
end


if TetGenIO then
  getmetatable(TetGenIO)["__call"]=callnew
end
]]

--
-- Assemble GridGenIO facet input lists from Lua face list.
--
-- void genfacets_f(GridGenIO *gg, table facets);

function genfacets_f(gg,facets)
   local npolypoints,npoly,nfacets,nfholes,nholes
   nfacets=table.getn(facets)
   local i
   npolypoints=0
   npoly=0
   nfholes=0
   nholes=0
   for i=1,nfacets do
      local nfacetpoly=table.getn(facets[i].polygons)
      npoly=npoly+nfacetpoly
      if facets[i].holepoints then nholes=nholes+table.getn(facets[i].holepoints) end
      for j=1,nfacetpoly do
	 npolypoints=npolypoints+table.getn(facets[i].polygons[j])
      end
   end

   print(npolypoints,npoly,nfacets,nholes)
   gg.facetpolygonlist=mbase.ia1Create(npolypoints)
   gg.facetpolygonindexes=mbase.ia1Create(npoly+1)
   gg.facetpolygons=mbase.ia1Create(nfacets+1)

   if nholes>0 then 
      gg.facetholepoints=mbase.da2Create(nholes,3) 
      gg.facetholes=mbase.ia1Create(nfacets+1) 
      gg.facetholes[1]=1
   end
   
   local polyidx=1
   local faceidx=1
   local ipoly=1
   local hidx=1
   
   gg.facetpolygonindexes[1]=1
   gg.facetpolygons[1]=1
   for i=1,nfacets do
      local nfacetpoly=table.getn(facets[i].polygons)
      faceidx=faceidx+nfacetpoly
      gg.facetpolygons[i+1]=faceidx
      for j=1,nfacetpoly do
	 for k=1,table.getn(facets[i].polygons[j]) do
	    gg.facetpolygonlist[polyidx]=facets[i].polygons[j][k]
	    polyidx=polyidx+1
	 end
	 gg.facetpolygonindexes[ipoly+1]=polyidx
	 ipoly=ipoly+1
      end
      if facets[i].holepoints then
	 for j=1,table.getn(facets[i].holepoints) do
	    for k=1,3 do gg.facetholepoints[hidx][k]=facets[i].holepoints[j][k] end
	    hidx=hidx+1
	 end
      end
      if nholes>0 then gg.facetholes[i+1]=hidx end
   end
end

-- assemble facet input lists from segment list
-- in the pre grid
function genfacets_s(gg,facets)
   local npolypoints,nfacets
   nfacets=table.getn(facets)
   npolypoints=0
   for i=1,nfacets do
      npolypoints=npolypoints+table.getn(facets[i])
   end

   gg.facetzones=IArray1(gg.segmentzones)
   gg.facetpolygonlist=mbase.ia1Create(npolypoints)
   gg.facetpolygonindexes=mbase.ia1Create(nfacets+1)
   gg.facetpolygons=mbase.ia1Create(nfacets+1)
   for i=1,nfacets+1 do gg.facetpolygons[i]=i end
   
   local polyidx=1
   gg.facetpolygonindexes[1]=1
   for i=1,nfacets do
      for k=1,table.getn(facets[i]) do
	 gg.facetpolygonlist[polyidx]=facets[i][k]
	 polyidx=polyidx+1
      end
      gg.facetpolygonindexes[i+1]=polyidx
   end
end


------------------------------------------------
---
--- \brief Set volume constraint function for ggVolume.
--- 
--- The Lua-function should have one parameter used for
--- the ggVolume object.
--- 
--- \param ggVolume and constraint function.
--+ ggVolumeSetConstraint( Ptr<ggVolume> vol, Lua-function )

function ggVolumeSetConstraint(vol,constraint)
  swigext_extend(vol,"constraint")
  vol.constraint=constraint
end

function ggRefineSetConstraint(vol,constraint)
  swigext_extend(vol,"constraint")
  vol.constraint=constraint
end

------------------------------------------------
---
--- \brief Set facetregions function for ggFacetRegions.
--- 
--- The Lua-function should have one parameter used for
--- the ggFacetRegions object.
--- 
--- \param ggFacetRegions and facetregion function.
--+ ggFacetRegionsSetFacetRegions( Ptr<ggFacetRegions> freg, Lua-function )

function ggFacetRegionsSetFacetRegions(freg,facetregion)
  swigext_extend(freg,"facetregion")
  freg.facetregion=facetregion
end


------------------------------------------------
---
--- \brief Set cellregions function for ggCellRegions.
--- 
--- The Lua-function should have one parameter used for
--- the ggCellRegions object.
--- 
--- \param ggCellRegions and cellregion function.
--+ ggCellRegionsSetCellRegions( Ptr<ggCellRegions> creg, Lua-function )

function ggCellRegionsSetCellRegions(creg,cellregion)
  swigext_extend(creg,"cellregion")
  creg.cellregion=cellregion
end


--------------------------------------------
-- dummy refinement function for triangle


using(mgridio)

